﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Web;
using System.Web.Mvc;
using System.Web.Mvc.Async;

using System.Threading;
using System.Net;
using System.IO;
using System.Collections.Specialized;
using System.Text;
using System.Text.RegularExpressions;

using HtmlAgilityPack;

namespace Elisy.AspNetProxy.Controllers
{
    [HandleError]
    public class ProxyController : AsyncController
    {
        public string Language { get; set; }
        public string Application { get; set; }

        protected override void Initialize(System.Web.Routing.RequestContext requestContext)
        {
            base.Initialize(requestContext);

            if (requestContext.RouteData.Values.ContainsKey("lang"))
                Language = requestContext.RouteData.Values["lang"].ToString();
            //else
            //    Language = "ru_RU";

            if (requestContext.RouteData.Values.ContainsKey("application"))
                Application = requestContext.RouteData.Values["application"].ToString();
            else
                RedirectToAction("Index", "Home");
        }

        public void TransferAsync(string pathInfo, string sysver)
        {
            AsyncManager.OutstandingOperations.Increment();

            ViewBag.SysVer = sysver;
            ViewBag.PathInfo = pathInfo;

            HttpWebRequest remoteRequest = (HttpWebRequest)HttpWebRequest.Create(new Uri("http://demo-ma.1c.ru/" + Application + (string.IsNullOrEmpty(Language)? "" : "/" + Language) + "/" + pathInfo + Request.Url.Query));
            remoteRequest.Method = Request.HttpMethod;
            remoteRequest.CookieContainer = new CookieContainer();
            if (Request.UrlReferrer != null)
                remoteRequest.Referer = Request.UrlReferrer.ToString();
            remoteRequest.UserAgent = Request.UserAgent;

            for (int i = 0; i < Request.Cookies.Count; i++) 
            {
                HttpCookie cookie = Request.Cookies.Get(i);

                Cookie newCookie = new Cookie();

                newCookie.Domain = remoteRequest.RequestUri.Host;

                newCookie.Expires  = cookie.Expires;
                newCookie.Name     = cookie.Name;
                newCookie.Path     = cookie.Path;
                newCookie.Secure   = cookie.Secure;
                newCookie.Value    = cookie.Value;

                remoteRequest.CookieContainer.Add(newCookie);
            }

            foreach(string key in Request.Headers)
            {
                if (key == "Connection")
                {
                    try
                    {
                        remoteRequest.Connection = Request.Headers.Get(key);
                    }
                    catch (Exception)
                    { }
                    continue;
                }
                if (key == "Accept")
                {
                    remoteRequest.Accept = Request.Headers.Get(key);
                    continue;
                }
                if (key == "Host")
                    continue;
                if (key == "User-Agent")
                    continue;
                if (key == "Referer")
                    continue;
                if (key == "Content-Length")
                    continue;
                if (key == "Content-Type")
                {
                    remoteRequest.ContentType = Request.Headers.Get(key);
                    continue;
                }
                remoteRequest.Headers.Add(key, Request.Headers.Get(key));
            }

            if (remoteRequest.Method == "POST")
            {
                using (var inputStream = remoteRequest.GetRequestStream())
                {
                    MemoryStream memoryStream = new MemoryStream();

                    byte[] buffer = new byte[255];
                    int bytesRead;
                    double totalBytesRead = 0;
                    Request.InputStream.Position = 0;
                    while ((bytesRead = Request.InputStream.Read(buffer, 0, buffer.Length)) > 0)
                    {
                        totalBytesRead += bytesRead;
                        memoryStream.Write(buffer, 0, bytesRead);
                    }

                    inputStream.Write(memoryStream.ToArray(), 0, (int)memoryStream.Length);
                    memoryStream.Close();
                }
            }

            remoteRequest.BeginGetResponse(result =>
                {
                    try
                    {
                        WebResponse response = remoteRequest.EndGetResponse(result);
                        AsyncManager.Parameters["response"] = (HttpWebResponse)response;
                    }
                    catch (WebException e)
                    {
                        AsyncManager.Parameters["response"] = (HttpWebResponse)e.Response;
                    }
                    AsyncManager.OutstandingOperations.Decrement();
                },
                null
            );


        }

        public ActionResult TransferCompleted(HttpWebResponse response)
        {
            using (var responseStream = response.GetResponseStream())
            {
                MemoryStream memoryStream = new MemoryStream();

                byte[] buffer = new byte[255];
                int bytesRead;
                double totalBytesRead = 0;
                while ((bytesRead = responseStream.Read(buffer, 0, buffer.Length)) > 0)
                {
                    totalBytesRead += bytesRead;
                    memoryStream.Write(buffer, 0, bytesRead);
                }
                ViewBag.ResponseContent = memoryStream.ToArray();
            }

            //Add new content
            if (ViewBag.PathInfo != null)
            {
                if (ViewBag.PathInfo == "mainform.html")
                {
                    HtmlDocument html = new HtmlDocument();
                    html.OptionFixNestedTags = true;
                    html.LoadHtml(Encoding.UTF8.GetString(ViewBag.ResponseContent, 0, ViewBag.ResponseContent.Length));

                    var res = html.DocumentNode.SelectSingleNode("//div[@id='preloader']");
                    HtmlNode node = html.CreateElement("img");
                    node.Attributes.Add("id", "1csoftware-powered");
                    node.Attributes.Add("style", "position:absolute;top:10px;right:10px;");
                    node.Attributes.Add("src", VirtualPathUtility.ToAbsolute("~/i/1csoftware.png"));
                    res.ChildNodes.Add(node);

                    ViewBag.ResponseContent = Encoding.UTF8.GetBytes(html.DocumentNode.OuterHtml);
                }
            }

            return new ContentActionResult() { RemoteResponse = response, FilePath = filePath, ResponseContent = ViewBag.ResponseContent };
        }

    }

    public class ContentActionResult : ActionResult
    {
        public HttpWebResponse RemoteResponse { get; set; }
        public string FilePath { get; set; }
        public byte[] ResponseContent { get; set; }

        public override void ExecuteResult(ControllerContext context)
        {
            var response = context.HttpContext.Response;
            response.ContentType = RemoteResponse.ContentType;
            response.Charset = RemoteResponse.CharacterSet;
            response.StatusCode = (int)RemoteResponse.StatusCode;

            for (int i = 0; i < RemoteResponse.Cookies.Count; i++)
            {
                Cookie cookie = RemoteResponse.Cookies[i];

                HttpCookie newCookie = new HttpCookie(cookie.Name);

                //newCookie.Domain = cookie.Domain;
                newCookie.Domain = context.HttpContext.Request.Url.Host;

                if (string.IsNullOrEmpty(newCookie.Domain))
                    newCookie.Domain = context.HttpContext.Request.Url.Host;
                //newCookie.Domain = "localhost";

                newCookie.Expires = cookie.Expires;
                newCookie.Name = cookie.Name;
                newCookie.Path = cookie.Path;
                newCookie.Secure = cookie.Secure;
                newCookie.Value = cookie.Value;

                response.SetCookie(newCookie);
            }

            
            foreach (string key in RemoteResponse.Headers.AllKeys)
            {
                response.AddHeader(key, RemoteResponse.Headers.Get(key));
            }

            response.BinaryWrite(ResponseContent);
        }

    }
}
